# LOG LEVELS
SILENT = 0
DEBUG = 10
INFO = 20
WARN = 30
ERROR = 40
FATAL = 50


NEW_PROCESS = 0
ENVIRONMENT = 1
ENGINE = 2
EXTERNAL = 3


class log:
    SILENT = SILENT
    DEBUG = DEBUG
    INFO = INFO
    WARN = WARN
    ERROR = ERROR
    FATAL = FATAL


# PROCESS
class process:
    #: Spawn the node/engine in a separate process.
    #: Allows parallelization, but increases communication overhead due to the (de)serialization of messages.
    NEW_PROCESS: int = NEW_PROCESS
    #: Spawn the node/engine in the process of the environment.
    ENVIRONMENT: int = ENVIRONMENT
    #: Spawn a node in the process of the engine.
    #: If an :class:`~eagerx.core.entities.EngineNode` requires direct access to the
    #: :attr:`~eagerx.core.entities.Engine.simulator`,
    #: :attr:`~eagerx.core.entities.EngineNode.config`, and
    #: :attr:`~eagerx.core.entities.EngineNode.engine_config`,
    #: it must be spawned in the same process as the engine.
    ENGINE: int = ENGINE
    #: Spawn the node/engine in a separate process. This process is not spawned by the environment.
    #: Instead, the user is responsible for running the executable script with the appropriate arguments.
    #: This allows nodes to run distributed.
    EXTERNAL: int = EXTERNAL


COMPATIBLE_DTYPES = [
    "float32",
    "float64",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "bool",
    "str",
]


class BackendException(Exception):
    """
    Base class for exceptions in backend routines
    """

    pass


class Unspecified(object):
    pass



from warnings import warn
import copy
from typing import Dict, Optional, Union, Type, TYPE_CHECKING, List
import gymnasium as gym
from yaml import dump
import numpy as np

import eagerx
from eagerx.core.view import SpecView
from eagerx.utils.utils import (
    replace_None,
    deepcopy,
)
from eagerx.utils.utils_sub import substitute_args


if TYPE_CHECKING:
    from eagerx.core.entities import Engine
    from eagerx import EngineGraph


class EntitySpec(object):
    def __init__(self, params):
        super(EntitySpec, self).__setattr__("_params", params)

    def __setattr__(self, name, value):
        raise AttributeError("You cannot set the new attributes to EntitySpec.")

    def __str__(self):
        return dump(self._params)

    @property
    @deepcopy
    def params(self):
        return self._params


class BackendSpec(EntitySpec):
    """A parameter specification that specifies how :class:`~eagerx.core.env.BaseEnv` should initialize the selected backend."""

    def initialize(self, spec_cls):
        pass

    @property
    def config(self) -> SpecView:
        """Provides an API to get/set the parameters to initialize.

        :return: (mutable) API to get/set parameters.
        """
        return SpecView(self, depth=["config"], unlocked=True)


class ProcessorSpec(EntitySpec):
    """A parameter specification that specifies how :class:`~eagerx.core.env.BaseEnv` should initialize the processor."""

    def initialize(self, spec_cls):
        pass

    @property
    def config(self) -> SpecView:
        """Provides an API to get/set the parameters to initialize.

        :return: (mutable) API to get/set parameters.
        """
        return SpecView(self, depth=[], unlocked=True)


class EngineStateSpec(EntitySpec):
    """A parameter specification that specifies how :class:`~eagerx.core.env.BaseEnv` should initialize the engine state."""

    def initialize(self, spec_cls):
        pass

    @property
    def config(self) -> SpecView:
        """Provides an API to get/set the parameters to initialize.

        :return: API to get/set parameters.
        """
        return SpecView(self, depth=[], unlocked=True)


class BaseNodeSpec(EntitySpec):
    def __init__(self, params):
        super().__init__(params)

    def _lookup(self, depth, unlocked=False):
        name = self._params["config"]["name"]
        return SpecView(self, depth=[depth], name=name, unlocked=unlocked)

    @property
    def config(self) -> SpecView:
        """Provides an API to set/get the parameters to initialize.

        The default parameters are:

        - .. py:attribute:: Spec.config.name: str

            User specified unique node name.

        - .. py:attribute:: Spec.config.rate: float

            Rate (Hz) at which the :func:`~eagerx.core.entities.Node.callback` is called.

        - .. py:attribute:: Spec.config.process: int = 0

            Process in which the node is launched. See :class:`~eagerx.core.constants.process` for all options.

        - .. py:attribute:: Spec.config.color: str = grey

            Specifies the color of logged messages & node color in the GUI.
            Check-out the termcolor documentation for the supported colors.

        - .. py:attribute:: Spec.config.print_mode: int = 1

            Specifies the different modes for printing: `{1: TERMCOLOR, 2: ROS}`.

        - .. py:attribute:: Spec.config.log_level: int = 30

            Specifies the log level for the engine: `{0: SILENT, 10: DEBUG, 20: INFO, 30: WARN, 40: ERROR, 50: FATAL}`

        :return: API to get/set parameters.
        """
        return self._lookup("config", unlocked=True)

    @property
    def inputs(self) -> SpecView:
        """Provides an API to set/get the parameters of registered :func:`eagerx.core.register.inputs`.

        The mutable parameters are:

        - .. py:attribute:: Spec.inputs.<name>.window: int = 1

           A non-negative number that specifies the number of messages to pass to the node's :func:`~eagerx.core.entities.Node.callback`.

           - *window* = 1: Only the last received input message.

           - *window* = *x* > 1: The trailing last *x* received input messages.

           - *window* = 0: All input messages received since the last call to the node's :func:`~eagerx.core.entities.Node.callback`.

           .. note:: With *window* = 0, the number of input messages may vary and can even be zero.

        - .. py:attribute:: Spec.inputs.<name>.processor: ProcessorSpec = None

            A processor that preprocesses the received input message before passing it
            to the node's :func:`~eagerx.core.entities.Node.callback`.

        - .. py:attribute:: Spec.inputs.<name>.space: dict = None

            This space defines the format of valid messages.

        - .. py:attribute:: Spec.inputs.<name>.delay: float = 0.0

            A non-negative simulated delay (seconds). This delay is ignored if
            :attr:`~eagerx.core.entities.Engine.simulate_delays` = True
            in the engine's :func:`~eagerx.core.entities.Engine.spec`.

        - .. py:attribute:: Spec.inputs.<name>.skip: bool = False

            Skip the dependency on this input during the first call to the node's :func:`~eagerx.core.entities.Node.callback`.
            May be necessary to ensure that the connected graph is directed and acyclic.

        :return: API to get/set parameters.
        """
        return self._lookup("inputs")

    @property
    def outputs(self) -> SpecView:
        """Provides an API to set/get the parameters of registered :func:`eagerx.core.register.outputs`.

        The mutable parameters are:

        - .. py:attribute:: Spec.outputs.<name>.processor: ProcessorSpec = None

            A processor that preprocesses the output message, returned by :func:`~eagerx.core.entities.Node.callback`,
            before publishing it.

        - .. py:attribute:: Spec.outputs.<name>.space: dict = None

            This space defines the format of valid messages.

        :return: API to get/set parameters.
        """
        return self._lookup("outputs")

    @property
    def states(self) -> SpecView:
        """Provides an API to set/get the parameters of registered :func:`eagerx.core.register.states`.

        The mutable parameters are:

        - .. py:attribute:: Spec.states.<name>.space: dict = None

            This space defines the format of valid messages.

        :return: API to get/set parameters.
        """
        return self._lookup("states")

    def initialize(self, spec_cls):
        import eagerx.core.register as register

        try:
            params = register.LOOKUP_TYPES[spec_cls.callback]
        except KeyError:
            if spec_cls.__name__ == "EnvNode":
                params = dict()
            else:
                raise

        if "targets" in params:
            from eagerx.core.entities import ResetNode

            assert issubclass(
                spec_cls, ResetNode
            ), "You can only have targets registered for nodes that inherit from the ResetNode baseclass."
            add_ft = True
        else:
            add_ft = False

        # Set default components
        for component, cnames in params.items():
            for cname, space in cnames.items():
                if component == "outputs":
                    if cname not in self.config.outputs:
                        self.config.outputs.append(cname)
                    mapping = dict(
                        rate="$(config rate)",
                        processor=None,
                        space=space,
                    )
                    # Add feedthrough entries for each output if node is a reset node (i.e. when it has a target)
                    if add_ft:
                        mapping_ft = dict(
                            delay=0.0,
                            window=1,
                            skip=False,
                            processor=None,
                            space=space,
                            address=None,
                        )
                        with self.feedthroughs as d:
                            d[cname] = mapping_ft
                elif component == "inputs":
                    if cname not in self.config.inputs:
                        self.config.inputs.append(cname)
                    space = eagerx.Space(shape=(), dtype="int64") if cname == "tick" else space
                    mapping = dict(
                        delay=0.0,
                        window=1,
                        skip=False,
                        processor=None,
                        space=space,
                        address=None,
                    )
                elif component == "targets":
                    if cname not in self.config.targets:
                        self.config.targets.append(cname)
                    mapping = dict(
                        processor=None,
                        space=space,
                        address=None,
                    )
                else:
                    if cname not in self.config.states:
                        self.config.states.append(cname)
                    component = "states"
                    mapping = dict(
                        processor=None,
                        space=space,
                    )
                with getattr(self, component) as d:
                    d[cname] = mapping

    def add_input(
        self,
        cname: str,
        window: int = 1,
        delay: float = 0.0,
        skip: bool = False,
        address: str = None,
        processor: Optional[ProcessorSpec] = None,
        space: Optional[gym.spaces.Space] = None,
    ):
        mapping = dict(
            window=window,
            delay=delay,
            skip=skip,
            space=space,
            address=address,
            processor=processor.params if processor else None,
        )
        with self.inputs as d:
            d[cname] = mapping

    def add_output(
        self,
        cname: str,
        processor: Optional[ProcessorSpec] = None,
        space: Optional[gym.spaces.Space] = None,
    ):
        mapping = dict(
            rate="$(config rate)",
            space=space,
            processor=processor.params if processor else None,
        )
        with self.outputs as d:
            d[cname] = mapping

    def build(self, ns: str):
        params = self.params  # Creates a deepcopy
        name = self.config.name
        entity_id = self.config.entity_id

        # Replace args in .yaml
        context = {
            "ns": {"env_name": ns, "node_name": name},
            "config": params["config"],
        }
        substitute_args(params, context, only=["config", "ns"])

        # Process inputs
        inputs = []
        for cname in self.config.inputs:
            assert (
                cname in params["inputs"]
            ), f'Received unknown {"input"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
            assert (
                "targets" not in params or cname not in params["targets"]
            ), f'Input "{cname}" cannot have the same cname as a target. Change either the input or target cname. Check the spec of "{name}" with entity_id "{entity_id}".'
            n = RxInput(name=cname, **params["inputs"][cname])
            inputs.append(n)

        # Process outputs
        outputs = []
        for cname in self.config.outputs:
            msg = f"The rate ({params['outputs'][cname]['rate']} Hz) set for action '{cname}' does not equal the environment rate ({self.config.rate} Hz)."
            assert params["outputs"][cname]["rate"] == self.config.rate, msg
            assert (
                cname in params["outputs"]
            ), f'Received unknown {"output"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
            if "address" in params["outputs"][cname]:
                address = params["outputs"][cname].pop("address")
            else:
                address = "%s/outputs/%s" % (name, cname)
            n = RxOutput(name=cname, address=address, **params["outputs"][cname])
            outputs.append(n)

        states = []
        for cname in self.config.states:
            assert (
                cname in params["states"]
            ), f'Received unknown {"state"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
            if "address" in params["states"][cname]:  # if 'env/supervisor', the state address is pre-defined (like an input)
                n = RxState(name=cname, **params["states"][cname])
            else:
                address = "%s/states/%s" % (name, cname)
                n = RxState(name=cname, address=address, **params["states"][cname])
            states.append(n)

        targets = []
        if "targets" in self.config:
            for cname in self.config.targets:
                assert (
                    cname in params["targets"]
                ), f'Received unknown {"target"} "{cname}". Check the spec of "{name}" with entity_id "{entity_id}".'
                n = RxState(name=cname, **params["targets"][cname])
                targets.append(n)

        feedthroughs = []
        if "feedthroughs" in params:
            assert "targets" in self.config, f'No targets defined for ResetNode "{name}".'
            assert len(self.config.targets) > 0, f'No targets selected for ResetNode "{name}".'
            for cname in self.config.outputs:
                # Add output details  to feedthroughs
                assert (
                    cname in params["feedthroughs"]
                ), f'Feedthrough "{cname}" must directly correspond to a selected output. Check the spec of "{name}" with entity_id "{entity_id}".'
                params["feedthroughs"][cname]["space"] = params["outputs"][cname]["space"]
                n = RxFeedthrough(feedthrough_to=cname, **params["feedthroughs"][cname])
                feedthroughs.append(n)

        params["outputs"] = [i.build(ns=ns) for i in outputs]
        params["inputs"] = [i.build(ns=ns) for i in inputs]
        params["states"] = [i.build(ns=ns) for i in states]
        params["targets"] = [i.build(ns=ns) for i in targets]
        params["feedthroughs"] = [i.build(ns=ns) for i in feedthroughs]

        # Create rate dictionary with outputs
        chars_ns = len(ns) + 1
        rate_dict = dict()
        for i in params["outputs"]:
            assert i["rate"] is not None and isinstance(i["rate"], (int, float)) and i["rate"] > 0, (
                f'The rate of node "{name}" (and output cname "{i["name"]}") is misspecified: rate="{i["rate"]}". '
                'Make sure that it is of type(rate)=("int", "float",) and rate > 0.'
            )
            address = i["address"][chars_ns:]
            rate_dict[address] = i["rate"]  # {'rate': i['rate']}

        # Put parameters in node namespace (watch out, order of dict keys probably matters...)
        node_params = {name: params, "rate": rate_dict}
        return replace_None(node_params)


class NodeSpec(BaseNodeSpec):
    """A parameter specification that specifies how :class:`~eagerx.core.env.BaseEnv` should initialize the node."""

    pass


class ResetNodeSpec(BaseNodeSpec):
    """A parameter specification that specifies how :class:`~eagerx.core.env.BaseEnv` should initialize the node."""

    @property
    def targets(self) -> SpecView:
        """Provides an API to set/get the parameters of registered :func:`eagerx.core.register.targets`.

        The mutable parameters are:

        - .. py:attribute:: Spec.targets.<name>.processor: ProcessorSpec = None

            A processor that preprocesses the received state message before passing it
            to the node's :func:`~eagerx.core.entities.ResetNode.callback`.

        :return: API to get/set parameters.
        """
        return self._lookup("targets")

    @property
    def feedthroughs(self) -> SpecView:
        """Provides an API to set/get the parameters of a feedthrough corresponding to registered :func:`eagerx.core.register.outputs`.

        The mutable parameters are:

        - .. py:attribute:: Spec.feedthroughs.<name>.processor: ProcessorSpec = None

            A processor that preprocesses the received input message before passing it
            to the node's :func:`~eagerx.core.entities.Node.callback`.

        - .. py:attribute:: Spec.feedthroughs.<name>.space: dict = None

            This space defines the format of valid messages.

        - .. py:attribute:: Spec.feedthroughs.<name>.delay: float = 0.0

            A non-negative simulated delay (seconds). This delay is ignored if
            :attr:`~eagerx.core.entities.Engine.simulate_delays` = True
            in the engine's :func:`~eagerx.core.entities.Engine.spec`.

        :return: API to get/set parameters.
        """
        return self._lookup("feedthroughs")


class ObjectSpec(EntitySpec):
    """A parameter specification of an object."""

    def __init__(self, params):
        super().__init__(params)

    def _lookup(self, depth, unlocked=False):
        name = self._params["config"]["name"]
        return SpecView(self, depth=[depth], name=name, unlocked=unlocked)

    def gui(
        self,
        engine_cls: Type["Engine"],
        interactive: Optional[bool] = True,
        resolution: Optional[List[int]] = None,
        filename: Optional[str] = None,
    ) -> Union[None, np.ndarray]:
        """Opens a graphical user interface of the object's engine implementation.

        .. note:: Requires `eagerx-gui`:

        .. highlight:: python
        .. code-block:: python

            pip3 install eagerx-gui

        :param engine_cls: The class engine (not instance!) that was used to register the engine implementation (e.g. "PybulletEngine").
        :param interactive: If `True`, an interactive application is launched.
                            Otherwise, an RGB render of the GUI is returned.
                            This could be useful when using a headless machine.
        :param resolution: Specifies the resolution of the returned render when `interactive` is `False`.
                           If `interactive` is `True`, this argument is ignored.
        :param filename: If provided, the GUI is rendered to an svg file with this name.
                         If `interactive` is `True`, this argument is ignored.
        :return: RGB render of the GUI if `interactive` is `False`.
        """
        import eagerx.core.register as register

        spec_copy = ObjectSpec(self.params)
        spec_copy._params["engine"] = {}
        engine = engine_cls.get_specification()
        graph = register.add_engine(spec_copy, engine)
        return graph.gui(interactive=interactive, resolution=resolution, filename=filename)

    @property
    def engine(self) -> Union[SpecView]:
        """Provides an API to set/get the parameters of an engine-specific implementation.

        The mutable parameters are:

        - Arguments (excluding spec) of the selected engine's :func:`~eagerx.core.entities.Engine.add_object` method.

        - .. py:attribute:: Spec.engine.states.<name>: EngineState

            Link an :class:`~eagerx.core.specs.EngineState` to a registered state with :func:`eagerx.core.register.states`.

        :return: API to get/set parameters.
        """
        return SpecView(self, depth=["engine"], name=self._params["config"]["name"])

    @property
    def sensors(self) -> SpecView:
        """Provides an API to set/get the parameters of registered :func:`eagerx.core.register.sensors`.

        The mutable parameters are:

        - .. py:attribute:: Spec.sensors.<name>.rate: float = 1.0

            Rate (Hz) at which the sensor's :func:`~eagerx.core.entities.EngineNode.callback` is called.

        - .. py:attribute:: Spec.sensors.<name>.space: dict = None

            This space defines the format of valid messages.

        :return: API to get/set parameters.
        """
        return self._lookup("sensors")

    @property
    def actuators(self) -> SpecView:
        """Provides an API to set/get the parameters of registered :func:`eagerx.core.register.actuators`.

        The mutable parameters are:

        - .. py:attribute:: Spec.actuators.<name>.rate: float = 1.0

            Rate (Hz) at which the actuator's :func:`~eagerx.core.entities.EngineNode.callback` is called.

        - .. py:attribute:: Spec.actuators.<name>.window: int = 1

           A non-negative number that specifies the number of messages to pass to the node's
           :func:`~eagerx.core.entities.EngineNode.callback`.

           - *window* = 1: Only the last received input message.

           - *window* = *x* > 1: The trailing last *x* received input messages.

           - *window* = 0: All input messages received since the last call to the node's
             :func:`~eagerx.core.entities.EngineNode.callback`.

           .. note:: With *window* = 0, the number of input messages may vary and can even be zero.


        - .. py:attribute:: Spec.actuators.<name>.space: dict = None

            This space defines the format of valid messages.

        - .. py:attribute:: Spec.actuators.<name>.delay: float = 0.0

            A non-negative simulated delay (seconds). This delay is ignored if
            :attr:`~eagerx.core.entities.Engine.simulate_delays` = True
            in the engine's :func:`~eagerx.core.entities.Engine.spec`.

        - .. py:attribute:: Spec.actuators.<name>.skip: bool = False

            Skip the dependency on this input during the first call to the node's :func:`~eagerx.core.entities.EngineNode.callback`.
            May be necessary to ensure that the connected graph is directed and acyclic.

        :return: API to get/set parameters.
        """
        return self._lookup("actuators")

    @property
    def states(self) -> SpecView:
        """Provides an API to set/get the parameters of registered :func:`eagerx.core.register.engine_states`.

        The mutable parameters are:

        - .. py:attribute:: Spec.states.<name>.space: dict = None

            This space defines the format of valid messages.

        :return: API to get/set parameters.
        """
        return self._lookup("states")

    @property
    def config(self) -> SpecView:
        """Provides an API to set/get the parameters to initialize.

        The default parameters are:

        - Additional parameters registered with the :func:`eagerx.core.register.config` decorator.

        - .. py:attribute:: Spec.config.name: str

            User specified unique object name.

        - .. py:attribute:: Spec.config.actuators: list

            List with selected actuators. Must be a subset of the registered :func:`eagerx.core.register.actuators`.

        - .. py:attribute:: Spec.config.sensors: list

            List with selected sensors. Must be a subset of the registered :func:`eagerx.core.register.sensors`.

        - .. py:attribute:: Spec.config.states: list

            List with selected engine_states. Must be a subset of the registered :func:`eagerx.core.register.engine_states`.

        :return: API to get/set parameters.
        """
        return self._lookup("config", unlocked=True)

    def initialize(self, spec_cls):
        import eagerx.core.register as register

        agnostic = register.LOOKUP_TYPES[spec_cls.make]

        # Set default components
        for component, cnames in agnostic.items():
            for cname, space in cnames.items():
                if component == "sensors":
                    mapping = dict(
                        rate=1,
                        processor=None,
                        space=space,
                    )
                elif component == "actuators":
                    mapping = dict(
                        rate=1,
                        delay=0.0,
                        window=1,
                        skip=False,
                        processor=None,
                        space=space,
                    )
                else:
                    component = "states"
                    mapping = dict(
                        processor=None,
                        space=space,
                    )
                with getattr(self, component) as d:
                    d[cname] = mapping
                # Select component per default
                if cname not in getattr(self.config, component):
                    getattr(self.config, component).append(cname)

    def _initialize_object_graph(self):
        mapping = dict()
        for component in ["sensors", "actuators"]:
            try:
                mapping[component] = getattr(self, component)
            except AttributeError:
                continue

        from eagerx.core.graph_engine import EngineGraph

        graph = EngineGraph.create(**mapping)
        return graph


class EngineSpec(BaseNodeSpec):
    """A parameter specification that specifies how :class:`~eagerx.core.env.BaseEnv` should initialize the engine."""

    @property
    def config(self) -> SpecView:
        """Provides an API to set/get the parameters to initialize.

        The default parameters are:

        - .. py:attribute:: Spec.config.rate: float

            Rate (Hz) at which the :func:`~eagerx.core.entities.Engine.callback` is called.

        - .. py:attribute:: Spec.config.process: int = 0

            Process in which the engine is launched. See :class:`~eagerx.core.constants.process` for all options.

        - .. py:attribute:: Spec.config.sync: bool = True

            Flag that specifies whether we run reactive or asynchronous.

        - .. py:attribute:: Spec.config.real_time_factor: float = 0

            A specified upper bound on the real-time factor. `Wall-clock-rate`=`real_time_factor`*`rate`.
            If `real_time_factor` < 1 the simulation is slower than real time.

        - .. py:attribute:: Spec.config.simulate_delays: bool = True

            Flag that specifies whether input delays are simulated.
            You probably want to set this to `False` when running in the real-world.

        - .. py:attribute:: Spec.config.color: str = grey

            Specifies the color of logged messages. Check-out the termcolor documentation for the supported colors.

        - .. py:attribute:: Spec.config.print_mode: int = 1

            Specifies the different modes for printing: `{1: TERMCOLOR, 2: ROS}`.

        - .. py:attribute:: Spec.config.log_level: int = 30

            Specifies the log level for the engine: `{0: SILENT, 10: DEBUG, 20: INFO, 30: WARN, 40: ERROR, 50: FATAL}`.

        :return: API to get/set parameters.
        """
        return self._lookup("config", unlocked=True)

    @property
    def objects(self) -> SpecView:
        """Provides an API to set/get the parameters to add an object to the engine.

        To add a new object, please use :func:`~eagerx.core.specs.Enginespec.add_object`.

        Arguments correspond to the signature of :func:`~eagerx.core.entities.Engine.add_object`.

        :return: API to get/set parameters.
        """
        return self._lookup("objects", unlocked=True)

    def add_object(
        self,
        name: str,
        **kwargs: Union[bool, int, float, str, List, Dict],
    ) -> None:
        """Adds an object to the simulator that is interfaced by the engine.

        :param kwargs: Other arguments of :func:`~eagerx.core.entities.Engine.add_object`.
        """
        # todo: check arguments of engine_spec.add_object(...) are arguments of engine.add_object(...).
        with self.objects as d:
            assert name not in d, f"There is already an object called `{name}` added. Names must be unique."
            d[name] = dict(engine_states=dict(), nodes=dict(), add_object=dict())
            with d[name].add_object as s:
                s["name"] = name
                s.update(kwargs)

    def build(self, ns: str):
        params = super().build(ns)

        # Convert engine states
        for name, obj in params[self.config.name]["objects"].items():
            for cname, view in obj["engine_states"].items():
                address = f"{name}/states/{cname}"
                processor = view["processor"] if view["processor"] != "null" else None
                spec = RxEngineState(cname, address, view["state"], processor, view["space"])
                obj["engine_states"][cname] = spec.build(ns=ns)
        return replace_None(params)

    def _add_engine_states(self, name: str, spec: ObjectSpec):
        # Pop states that were not implemented.
        assert name in self.objects, f"There is no Object called `{name}' in engine.objects. First add the Object."
        states = spec.engine.states
        for cname in list(states.keys()):
            if cname in spec.config.states:
                if states[cname] is not None:
                    self._add_engine_state(
                        name, cname, states[cname], spec.states[cname]["space"], spec.states[cname]["processor"]
                    )
                else:
                    warn(
                        f"Engine state `{cname}` for object `{name}` will be ignored. "
                        f"There is no implementation provided for it in Engine `{self.config.entity_id}`."
                    )

    def _add_engine_state(self, name, cname, engine_state, space, processor=None):
        with self.objects[name].engine_states as s:
            s[cname] = dict(state=None, space=None, processor=None)
            s[cname]["state"] = engine_state
            s[cname]["space"] = space
            s[cname]["processor"] = processor

    def _initialize_engine_config(self, spec, engine_config):
        # Add default config
        with spec.engine as d:
            d.update(engine_config)
            d["name"] = spec.config.name
            d["states"] = {}
            # Add all states to engine-specific params
            with d.states as s:
                for cname in spec.states.keys():
                    s[cname] = None

    def _register_object(self, spec: ObjectSpec) -> "EngineGraph":
        spec = copy.deepcopy(spec)

        # Construct context & replace placeholders
        context = {"config": spec.config.to_dict()}
        substitute_args(spec._params["config"], context, only=["config"])  # First resolve args within the context
        substitute_args(spec._params, context, only=["config"])  # Resolve rest of params

        # Add engine entry
        import eagerx.core.register as register

        spec._params["engine"] = {}
        return register.add_engine(spec, self)


# REQUIRED FOR BUILDING SPECS


class Component(object):
    def __init__(self, **kwargs):
        # Iterates over provided arguments and sets the provided arguments as class properties
        for key, value in kwargs.items():
            if key == "__class__":
                continue  # Skip if __class__ type
            setattr(self, key, value)


class RxInput(Component):
    def __init__(
        self,
        name: str,
        address: str,
        window: int = 0,
        processor: Dict = None,
        space: Dict = None,
        delay: float = 0.0,
        skip: bool = False,
        dtype: str = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxInput, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        # Set dtype if not already set by source
        if params["dtype"] is None:
            params["dtype"] = params["space"]["dtype"]
        return params


class RxOutput(Component):
    def __init__(
        self,
        name: str,
        address: str,
        rate: float,
        processor: Dict = None,
        space: Dict = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxOutput, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        params["dtype"] = params["space"]["dtype"]
        return params


class RxFeedthrough(Component):
    def __init__(
        self,
        address: str,
        feedthrough_to: str,
        window: int = 1,
        processor: Dict = None,
        space: Dict = None,
        delay: float = 0.0,
        skip: bool = False,
        dtype: str = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxFeedthrough, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        # Set dtype if not already set by source
        if params["dtype"] is None:
            params["dtype"] = params["space"]["dtype"]
        return params


class RxState(Component):
    def __init__(
        self,
        name: str,
        address: str,
        space: Dict,
        processor: Dict = None,
        dtype: str = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxState, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        # Set dtype if not already set by source
        if params["dtype"] is None:
            params["dtype"] = params["space"]["dtype"]
        return params


class RxEngineState(Component):
    def __init__(
        self,
        name: str,
        address: str,
        state: Dict,
        processor: Dict = None,
        space: Dict = None,
    ):
        # Store parameters as properties in baseclass
        # IMPORTANT! Do not define variables locally you do **not** want to store
        # on the parameter server anywhere before calling the baseclass' constructor.
        kwargs = locals().copy()
        kwargs.pop("self")
        super(RxEngineState, self).__init__(**kwargs)

        # Calculate other parameters based on previously defined attributes.

        # Error check the parameters here.

    def build(self, ns=""):
        params = self.__dict__.copy()
        params["address"] = "/".join(filter(None, [ns, params["address"]]))
        params["dtype"] = params["space"]["dtype"]
        return params



# ROS1
try:
    import rospy
    import rosparam
    import rosgraph
    import rosservice
    from std_srvs.srv import Trigger, TriggerResponse, TriggerRequest
    import std_msgs.msg
except ImportError as e:
    raise ImportError(f"{e}. (HINT: you need to install ROS Melodic or Noetic and have it sourced.")

# OTHER
import logging
import typing
import time
import numpy as np
import atexit
from functools import wraps

# EAGERx
import eagerx
from eagerx.utils.utils import Header
from eagerx.core.pubsub import Publisher, Subscriber, ShutdownService
from eagerx.core.constants import (
    SILENT,
    DEBUG,
    INFO,
    ERROR,
    WARN,
    FATAL,
    COMPATIBLE_DTYPES,
    BackendException,
    Unspecified,
)

# A singleton that is used to check if an argument was specified.
_unspecified = Unspecified()


class Ros1(eagerx.Backend):

    BACKEND = "ROS1"
    DISTRIBUTED_SUPPORT = True
    MULTIPROCESSING_SUPPORT = True
    COLAB_SUPPORT = True

    @classmethod
    def make(cls, log_level=WARN) -> eagerx.specs.BackendSpec:
        spec = cls.get_specification()
        spec.config.log_level = log_level if isinstance(log_level, str) else eagerx.get_log_level()
        return spec

    def initialize(self):
        self._bnd = self
        _initialize("eagerx_backend_ros1", anonymous=True, log_level=INFO)

        # Set log level
        logger = logging.getLogger("rosout")
        logger.setLevel(self.log_level)

    def Publisher(self, address: str, dtype: str):
        return _Publisher(self._bnd, address, dtype)

    def Subscriber(self, address: str, dtype: str, callback, header: bool = False, callback_args=tuple()):
        return _Subscriber(self._bnd, address, dtype, callback, header, callback_args=callback_args)

    @staticmethod
    def logdebug(msg, *args, **kwargs):
        return rospy.logdebug(msg)

    @staticmethod
    def loginfo(msg, *args, **kwargs):
        return rospy.loginfo(msg)

    @staticmethod
    def logwarn(msg, *args, **kwargs):
        return rospy.logwarn(msg)

    @staticmethod
    def logerr(msg, *args, **kwargs):
        return rospy.logerr(msg)

    def register_environment(self, name: str, force_start: bool, fn: typing.Callable):
        # Check if there already exists an environment
        services = rosservice.get_service_list()
        if f"{name}/environment/shutdown" in services:
            if force_start:
                Ros1.logwarn(f"There already exists an environment named '{name}'. Shutting down existing environment.")
                shutdown_client = rospy.ServiceProxy(f"{name}/environment/shutdown", Trigger)
                shutdown_client.wait_for_service(1)
                shutdown_client(TriggerRequest())
            else:
                msg = (
                    f"There already exists an environment named '{name}'. Exiting now. Set 'force_start=True' if you "
                    "want to shutdown the existing environment."
                )
                Ros1.logerr(msg)
                raise BackendException

        # Setup remote shutdown procedure for environment
        def _remote_shutdown(req: TriggerRequest):
            Ros1.loginfo("Remote shutdown procedure started.")
            fn()
            Ros1.loginfo("Remote shutdown procedure completed.")
            return TriggerResponse(success=True, message="")

        shutdown_srv = rospy.Service(f"{name}/environment/shutdown", Trigger, _remote_shutdown)
        return _ShutdownService(shutdown_srv)

    def delete_param(self, param: str, level: int = 1) -> None:
        try:
            rosparam.delete_param(param)
            Ros1.loginfo(f'Parameters under namespace "{param}" deleted.')
        except rosgraph.masterapi.ROSMasterException as e:
            if level == 0:
                raise BackendException(e)
            elif level == 1:
                Ros1.logwarn(e)
            else:
                pass

    def upload_params(self, ns: str, values: typing.Dict, verbose: bool = False) -> None:
        return rosparam.upload_params(ns, values, verbose=verbose)

    def get_param(self, name: str, default: typing.Any = _unspecified):
        try:
            if isinstance(default, Unspecified):
                return rospy.get_param(name)
            else:
                return rospy.get_param(name, default=default)
        except rosgraph.masterapi.Error as e:
            raise BackendException(e)

    def spin(self):
        return rospy.spin()


# Private
_LOG_FNS = {
    SILENT: lambda print_str: None,
    DEBUG: rospy.logdebug,
    INFO: rospy.loginfo,
    WARN: rospy.logwarn,
    ERROR: rospy.logerr,
    FATAL: rospy.logfatal,
}

_LOG_LEVELS = {
    SILENT: rospy.DEBUG,
    DEBUG: rospy.DEBUG,
    INFO: rospy.INFO,
    WARN: rospy.WARN,
    ERROR: rospy.ERROR,
    FATAL: rospy.FATAL,
}

_NUMPY_TO_ROS = {key: None for key in COMPATIBLE_DTYPES}
_NUMPY_TO_ROS["str"] = std_msgs.msg.UInt8MultiArray
_NUMPY_TO_ROS["bool"] = std_msgs.msg.Int8MultiArray
_NUMPY_TO_ROS["float32"] = std_msgs.msg.Float32MultiArray
_NUMPY_TO_ROS["float64"] = std_msgs.msg.Float64MultiArray
_NUMPY_TO_ROS["int8"] = std_msgs.msg.Int8MultiArray
_NUMPY_TO_ROS["int16"] = std_msgs.msg.Int16MultiArray
_NUMPY_TO_ROS["int32"] = std_msgs.msg.Int32MultiArray
_NUMPY_TO_ROS["int64"] = std_msgs.msg.Int64MultiArray
_NUMPY_TO_ROS["uint8"] = std_msgs.msg.UInt8MultiArray
_NUMPY_TO_ROS["uint16"] = std_msgs.msg.UInt16MultiArray
_NUMPY_TO_ROS["uint32"] = std_msgs.msg.UInt32MultiArray
_NUMPY_TO_ROS["uint64"] = std_msgs.msg.UInt64MultiArray
_ROS_TO_NUMPY = {val: key for key, val in _NUMPY_TO_ROS.items()}

_UNSUPPORTED = [key for key, value in _NUMPY_TO_ROS.items() if value is None]


def _is_roscore_running():
    return rosgraph.is_master_online()


def _launch_roscore():
    if not _is_roscore_running():

        import roslaunch

        uuid = roslaunch.rlutil.get_or_generate_uuid(options_runid=None, options_wait_for_master=False)
        roslaunch.configure_logging(uuid)
        roscore = roslaunch.parent.ROSLaunchParent(uuid, roslaunch_files=[], is_core=True)

        try:
            roscore.start()
        except roslaunch.core.RLException:
            Ros1.loginfo(
                "Roscore cannot run as another roscore/master is already running. Continuing without re-initializing the roscore."
            )
            # roscore = None
            raise  # todo: maybe not raise here?
    else:
        Ros1.loginfo(
            "Roscore cannot run as another roscore/master is already running. Continuing without re-initializing the roscore."
        )
        roscore = None
    return roscore


@wraps(rospy.init_node)
def _initialize(*args, log_level=INFO, **kwargs):
    roscore = _launch_roscore()  # First launch roscore (if not already running)
    try:
        rospy.init_node(*args, log_level=_LOG_LEVELS[log_level], **kwargs)
    except rospy.exceptions.ROSException as e:
        Ros1.logwarn(e)
    if roscore:
        atexit.register(roscore.shutdown)
    return roscore


def _dtype_to_ros1_msg_type(dtype: str):
    try:
        msg_type_cls = _NUMPY_TO_ROS[dtype]
        return msg_type_cls
    except KeyError:
        msg = f"Invalid dtype provided that is not supported ({dtype}). Supported dtypes are ({_NUMPY_TO_ROS.keys()})"
        raise KeyError(msg)


# Public Publisher/Subscriber functions


class _Publisher(Publisher):
    def __init__(self, backend: eagerx.Backend, address: str, dtype: str):
        super().__init__(backend)
        self._address = address
        self._dtype = dtype
        self._msg_type = _dtype_to_ros1_msg_type(dtype)
        self._pub = rospy.Publisher(address, self._msg_type, queue_size=0, latch=True)
        self._name = f"{self._address}"

    def _publish(self, msg: typing.Union[float, bool, int, str, np.ndarray, np.number], header: Header) -> None:
        # Convert python native types to numpy arrays.
        if isinstance(msg, float):
            msg = np.array(msg, dtype="float32")
        elif isinstance(msg, bool):
            msg = np.array(msg, dtype="bool")
        elif isinstance(msg, int):
            msg = np.array(msg, dtype="int64")

        # Check if message complies with space
        if not isinstance(msg, (np.ndarray, np.number, str)):
            Ros1.logerr(f"[publisher][{self._name}]: type(recv)={type(msg)}")
            time.sleep(10000000)

        # Check if dtypes of processed message matches outgoing message_type
        msg_dtype = msg.dtype.name if isinstance(msg, (np.ndarray, np.number)) else type(msg).__name__
        msg_type = _dtype_to_ros1_msg_type(msg_dtype)
        if not msg_type == self._msg_type:
            raise AssertionError(
                f"[publisher][{self._name}]: "
                f"(processed) message dtype ({msg_dtype}) corresponds to msg_type ({msg_type} which is incompatible"
                f" with the registered dtype ({self._dtype}) that corresponds to msg_type ({self._msg_type})."
            )

        # Convert to ROS messages
        shape = msg.shape if isinstance(msg, (np.ndarray, np.number)) else ()
        if msg_dtype == "uint8":
            msg = self._msg_type(data=msg.tobytes("C"))
        elif msg_dtype == "bool":
            msg = self._msg_type(data=[int(msg.item())])
        elif msg_dtype == "str":
            msg = self._msg_type(data=msg.encode("utf-8"))
        else:
            msg = self._msg_type(data=msg.reshape(-1))

        # HACK! Abuse MultiArrayDimension to encode header information
        seq, sc, wc = header
        secs, nsecs = self._backend.serialize_time(sc)
        sc_header = std_msgs.msg.MultiArrayDimension(label=str(seq), size=secs, stride=nsecs)
        msg.layout.dim.append(sc_header)
        secs, nsecs = self._backend.serialize_time(wc)
        wc_header = std_msgs.msg.MultiArrayDimension(label=str(seq), size=secs, stride=nsecs)
        msg.layout.dim.append(wc_header)

        # Add dimension information
        for i, d in enumerate(shape):
            dim = std_msgs.msg.MultiArrayDimension(label=str(i), size=d, stride=sum(shape[i:]))
            msg.layout.dim.append(dim)
        # if self._address == "/rx/obj/states/N9/set":
        #     print(f"[publisher][{self._name}]: N9 sent!")
        # elif self._address == "/rx/obj/states/N9/done":
        #     print(f"[publisher][{self._name}]: N9/done sent!")
        return self._pub.publish(msg)

    def unregister(self):
        return self._pub.unregister()


class _Subscriber(Subscriber):
    def __init__(self, backend: eagerx.Backend, address: str, dtype: str, callback, header: bool, callback_args=tuple()):
        super().__init__(backend, header)
        self._address = address
        self._dtype = dtype
        self._msg_type = _dtype_to_ros1_msg_type(dtype)
        self._cb_wrapped = callback
        self._cb_args = callback_args
        self._sub = rospy.Subscriber(address, self._msg_type, callback=self.callback)
        self._name = f"{self._address}"

    def callback(self, msg):
        if not isinstance(msg, tuple(_ROS_TO_NUMPY)):
            Ros1.logerr(f"[subscriber][{self._name}]: type(recv)={type(msg)}")
            time.sleep(10000000)

        if self._header:
            seq = int(msg.layout.dim[0].label)
            sc_secs, sc_nsecs = msg.layout.dim[0].size, msg.layout.dim[0].stride
            wc_secs, wc_nsecs = msg.layout.dim[1].size, msg.layout.dim[1].stride
            sc, wc = self._backend.deserialize_time(sc_secs, sc_nsecs), self._backend.deserialize_time(wc_secs, wc_nsecs)
            header = Header(seq=seq, sc=sc, wc=wc)

        # Convert input message to numpy array
        if isinstance(msg, std_msgs.msg.Int8MultiArray) and self._dtype == "bool":
            msg = bool(msg.data[0])
        elif isinstance(msg, std_msgs.msg.UInt8MultiArray):
            if self._dtype == "str":
                msg = msg.data.decode("utf-8")
            else:
                s = [d.size for d in msg.layout.dim[2:]]
                msg = np.frombuffer(bytes(msg.data), dtype="uint8").reshape(s)
        else:
            s = [d.size for d in msg.layout.dim[2:]]
            msg = np.array(msg.data, dtype=_ROS_TO_NUMPY[type(msg)]).reshape(s)

        try:
            # if self._address == "/Pendulum_2_True_0/env/supervisor/end_reset":
            # if self._address == "/rx/obj/states/N9/set":
            #     print(f"[subscriber][{self._name}]: N9 received={msg}!")
            # elif self._address == "/rx/obj/states/N9/done":
            #     print(f"[subscriber][{self._name}]: N9/done received={msg}!")
            # Extract header information
            self._cb_wrapped(msg, header, *self._cb_args) if self._header else self._cb_wrapped(msg, *self._cb_args)
        except rospy.exceptions.ROSException as e:
            self.unregister()
            Ros1.logdebug(f"[subscriber][{self._name}]: Unregistered this subscription because of exception: {e}")

    def unregister(self):
        return self._sub.unregister()


class _ShutdownService(ShutdownService):
    def __init__(self, srv):
        self._srv = srv

    def unregister(self):
        self._srv.shutdown()


assert len(_UNSUPPORTED) == 0, f"Backend `{Ros1.BACKEND}` does not have support for the following dtypes: {_UNSUPPORTED}."



# RX IMPORTS
import rx
from rx import Observable, typing, operators as ops
from rx.disposable import Disposable, SingleAssignmentDisposable, CompositeDisposable
from rx.subject import Subject, BehaviorSubject
from rx.internal.concurrency import synchronized

# EAGERX IMPORTS
import eagerx.utils.utils
from eagerx.core.constants import (  # noqa
    SILENT,
    DEBUG,
    INFO,
    ERROR,
    WARN,
    FATAL,
)
from eagerx.utils.utils import (
    Info,
    Msg,
    Stamp,
)

# OTHER IMPORTS
import time
from math import floor
from collections import deque
from termcolor import colored
import datetime
import traceback
from os import getpid
from threading import current_thread, RLock
from typing import Callable, Optional, List, Any
import numpy as np


def cb_ft(cb_input, sync):
    # Fill output msg with number of node ticks
    output_msgs = dict()
    for key, msg in cb_input.items():
        if key not in ["node_tick", "t_n"]:
            if len(msg.msgs) > 0:
                output_msgs[key] = msg.msgs[-1]
            else:
                assert not sync, "Actions must always be fed through if we are running reactively."
                output_msgs[key] = None
    return output_msgs


def print_info(
    node_name,
    color,
    id=None,
    trace_type=None,
    value=None,
    date=None,
    log_level=DEBUG,
):
    msg = ""
    if date:
        msg += f"[{str(date)[:40].ljust(20)}]"
    # Add process ID
    msg += f"[{str(getpid())[:5].ljust(5)}]"
    # Add thread ID
    msg += f"[{current_thread().name.split('/')[-1][:15].ljust(15)}]"
    msg += f"[{node_name.split('/')[-1][:12].ljust(12)}]"
    if id:
        msg += f"[{id.split('/')[-1][:12].ljust(12)}]"
    msg += f" {trace_type}: {value}\n"
    print(colored(msg, color), end="")


def spy(id: str, node, log_level: int = DEBUG, mapper: Callable = lambda msg: msg):
    node_name = node.ns_name
    color = node.color

    effective_log_level = node.backend.log_level

    def _spy(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):

                if node.log_level >= effective_log_level and log_level >= effective_log_level:
                    print_info(
                        node_name,
                        color,
                        id,
                        trace_type="",
                        value=str(mapper(value)),
                        log_level=log_level,
                    )
                observer.on_next(value)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _spy


def trace_observable(
    id: str,
    node,
    trace_next=False,
    trace_next_payload=False,
    trace_subscribe=False,
    date=None,
):  # pragma: no cover
    node_name = node.ns_name
    color = node.color

    def _trace(source):
        def on_subscribe(observer, scheduler):
            def on_next(value):
                if trace_next is True:
                    if trace_next_payload is True:
                        print_info(
                            node_name,
                            color,
                            id,
                            "on_next",
                            value,
                            date=date or datetime.datetime.now(),
                            log_level=DEBUG,
                        )
                    else:
                        print_info(
                            node_name,
                            color,
                            id,
                            "on_next",
                            "",
                            date=date or datetime.datetime.now(),
                            log_level=DEBUG,
                        )
                observer.on_next(value)

            def on_completed():
                value = ""
                print_info(
                    node_name,
                    color,
                    id,
                    "on_completed",
                    value,
                    date=date or datetime.datetime.now(),
                    log_level=DEBUG,
                )
                observer.on_completed()

            def on_error(error):
                if isinstance(error, Exception):
                    error_traceback = "%s, %s" % (
                        error,
                        traceback.print_tb(error.__traceback__),
                    )
                    print_info(
                        node_name,
                        color,
                        id,
                        "on_error",
                        error_traceback,
                        date=date or datetime.datetime.now(),
                    )
                else:
                    print_info(
                        node_name,
                        color,
                        id,
                        "on_error",
                        error,
                        date=date or datetime.datetime.now(),
                        log_level=ERROR,
                    )
                observer.on_error(error)

            def dispose():
                if trace_subscribe is True:
                    value = ""
                    print_info(
                        node_name,
                        color,
                        id,
                        "dispose",
                        value,
                        date=date or datetime.datetime.now(),
                        log_level=DEBUG,
                    )
                disposable.dispose()

            if trace_subscribe is True:
                value = ""
                print_info(
                    node_name,
                    color,
                    id,
                    "on_subscribe",
                    value,
                    date=date or datetime.datetime.now(),
                    log_level=DEBUG,
                )
            disposable = source.subscribe(
                on_next=on_next,
                on_error=on_error,
                on_completed=on_completed,
            )
            return Disposable(dispose)

        return rx.create(on_subscribe)

    return _trace


def flag_dict(name):
    def _init_flag_dict(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                flag_dict = {name: value}
                observer.on_next(flag_dict)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _init_flag_dict


def filter_dict():
    def _filter_dict(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                d = dict()
                for node_name, flag in value.items():
                    if flag is True:
                        continue
                    d[node_name] = flag
                observer.on_next(d)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _filter_dict


def switch_to_reset():
    def _switch_to_reset(source):
        def subscribe(observer, scheduler=None):
            reset_mode = [False]

            def on_next(value):
                if isinstance(value, bool):
                    reset_mode[0] = True  # if we receive a reset flag, turn on reset mode
                    observer.on_next(value)
                else:
                    if not reset_mode[0]:  # Check if we haven't previously received a reset message
                        observer.on_next(value)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _switch_to_reset


def create_msg_tuple(name: str, node_tick: int, msg: List[Any], stamp: List[Stamp], done: bool = None):
    info = Info(name=name, node_tick=node_tick, t_in=stamp, done=done)
    return Msg(info, msg)


def remap_state(name, sync, real_time_factor):
    def _remap_state(source):
        def subscribe(observer, scheduler=None):
            start = time.time()
            seq = [0]

            def on_next(value):
                node_tick = value[0][0]
                msg = value[0][1]
                done = value[1]
                wc = time.time()
                if sync:
                    sc = None
                else:
                    sc = (wc - start) / real_time_factor
                stamp = Stamp(seq[0], sc, wc)
                res = create_msg_tuple(name, node_tick, [msg], [stamp], done=done)
                seq[0] += 1
                observer.on_next(res)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _remap_state


def remap_target(name, sync, real_time_factor):
    def _remap_target(source):
        def subscribe(observer, scheduler=None):
            start = time.time()
            seq = [0]

            def on_next(value):
                node_tick = value[0]
                msg = value[1]
                wc = time.time()
                if sync:
                    sc = None
                else:
                    sc = (wc - start) / real_time_factor
                stamp = Stamp(seq[0], sc, wc)
                res = create_msg_tuple(name, node_tick, [msg], [stamp])
                seq[0] += 1
                observer.on_next(res)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _remap_target


def filter_info_for_printing(info):
    info_dict = dict()
    if info.rate_in:
        info_dict["rate_in"] = info.rate_in
    info_dict["t_in"] = [t.sc for t in info.t_in]
    if info.t_node:
        info_dict["t_node"] = [t.sc for t in info.t_node]
    if info.done:
        info_dict["done"] = info.done
    return info_dict


def remap_cb_input(mode=0):
    def _remap_cb_input(value):
        # mode=0 (info only), mode=1 (msgs only), mode=2 (all)
        if mode == 2:
            return value
        if isinstance(value, tuple):
            mapped_value = tuple([value[0].copy(), value[1].copy()])
            for i in mapped_value:
                for key, msg in i.items():
                    if key not in ["node_tick", "t_n"]:
                        if mode == 0:
                            i[key] = filter_info_for_printing(msg.info)
                        else:
                            i[key] = msg.msgs
        else:
            mapped_value = value.copy()
            for key, msg in mapped_value.items():
                if key not in ["node_tick", "t_n"]:
                    if mode == 0:
                        mapped_value[key] = filter_info_for_printing(msg.info)
                    else:
                        mapped_value[key] = msg.msgs
        return mapped_value

    return _remap_cb_input


def regroup_inputs(node, rate_node=1, is_input=True, perform_checks=True):
    node_name = node.ns_name
    color = node.color

    def _regroup_inputs(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                # Regroups all inputs into a single dict
                if is_input:
                    node_tick = value[0].info.node_tick
                    t_n = round(node_tick / rate_node, 12)
                    res = dict(node_tick=node_tick, t_n=t_n)
                else:
                    res = dict()
                for msg in value:
                    res[msg.info.name] = msg

                # Perform checks
                if perform_checks and is_input:
                    node_ticks = []
                    for msg in value:
                        node_ticks.append(msg.info.node_tick)
                    if len(node_ticks) > 0:
                        if not len(set(node_ticks)) == 1:
                            print_info(
                                node_name,
                                color,
                                "regroup_inputs",
                                trace_type="",
                                value="Not all node_ticks are the same: %s" % str(value),
                                log_level=ERROR,
                            )

                # Send regrouped input
                observer.on_next(res)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _regroup_inputs


def expected_inputs(N_out, rate_in, rate_out, delay, skip: bool):
    # In case of skip=True and the output rate is an exact multiple of the input rate, we need to add a slight offset
    eps = 1e-9

    # Constants
    offset = int(skip) * (int((rate_out - eps) / rate_in) if rate_out > rate_in else -1)

    # Alternative numerically unstable implementation
    # dt_out, dt_in = 1 / rate_out, 1 / rate_in
    # t_prev = dt_out * (N_out - 1 + offset) - delay
    # t = dt_out * (N_out + offset) - delay
    # N_in_prev = int(t_prev / dt_in)
    # N_in = int(t / dt_in)
    # delta = N_in - N_in_prev
    # j = (delta - 1 + int(not skip))
    # T = dt_out * N_out - delay - j * dt_in
    # correction = ceil(-T / dt_in)

    # Numerically stable implementation
    N_in_prev = int((rate_in * (N_out + offset - 1) - rate_out * rate_in * delay) // rate_out)
    N_in = int((rate_in * (N_out + offset) - rate_out * rate_in * delay) // rate_out)

    # Alternative (iterative) delay correction
    # delta = N_in - N_in_prev
    # for i in range(int(not skip), delta + int(not skip)):
    #     # Alternative numerically unstable implementation
    #     # t_in_delayed = dt_out * N_out - delay - i * dt_in
    #     # Numerically stable implementation
    #     t_in_delayed = (N_out * rate_in - delay * rate_out * rate_in - i * rate_out) / rate_in
    #     if t_in_delayed < 0:
    #         delta -= 1

    # Alternative delay correction
    delta = N_in - N_in_prev
    j = delta - 1 + int(not skip)
    # Numerically stable implementation
    T = (rate_in * N_out - rate_out * rate_in * delay - rate_out * j) / (rate_out * rate_in)
    correction = -floor(T * rate_in)
    corrected = delta - min(delta, max(0, correction))  # limits as follows: 0 < correction < delta

    # Overwrite t=0 dependencies, because there are none when skipping.
    num_est = corrected if N_out > 0 else int(not skip)
    return num_est


def generate_msgs(
    source_Nc: Observable,
    rate_node: float,
    name: str,
    rate_in: float,
    params: dict,
    sync: bool,
    real_time_factor: float,
    simulate_delays: bool,
    node=None,
):
    dt_i = 1 / rate_in

    def _generate_msgs(source_msg: Observable):
        window = params["window"]
        skip = int(params["skip"])

        def subscribe(observer: typing.Observer, scheduler: Optional[typing.Scheduler] = None) -> CompositeDisposable:
            start = time.time()
            msgs_queue: List = []
            t_i_queue: List = []
            num_queue: List = []
            tick_queue: List = []
            msgs_window = deque(maxlen=window)
            t_i_window = deque(maxlen=window)
            t_n_window = deque(maxlen=window)
            lock = RLock()

            @synchronized(lock)
            def next(i):
                if len(tick_queue) > 0:
                    if not sync or len(msgs_queue) >= num_queue[0]:
                        try:
                            tick = tick_queue.pop(0)
                            if sync:
                                # determine num_msgs
                                num_msgs = num_queue.pop(0)
                                msgs = msgs_queue[:num_msgs]
                                t_i = t_i_queue[:num_msgs]
                                msgs_queue[:] = msgs_queue[num_msgs:]
                                t_i_queue[:] = t_i_queue[num_msgs:]
                            else:  # Empty complete buffer
                                msgs = msgs_queue.copy()
                                t_i = t_i_queue.copy()
                                msgs_queue[:] = []
                                t_i_queue[:] = []
                        except Exception as ex:  # pylint: disable=broad-except
                            observer.on_error(ex)
                            return

                        # Determine t_n stamp
                        wc = time.time()
                        seq = tick
                        if sync:
                            sc = round(tick / rate_node, 12)
                        else:
                            sc = (wc - start) / real_time_factor
                        t_n = Stamp(seq, sc, wc)

                        if window > 0:
                            msgs_window.extend(msgs)
                            t_i_window.extend(t_i)
                            t_n_window.extend([t_n] * len(msgs))
                            wmsgs = list(msgs_window)
                            wt_i = list(t_i_window)
                            wt_n = list(t_n_window)
                        else:
                            wmsgs = msgs
                            wt_i = t_i
                            wt_n = [t_n] * len(msgs)
                        res = Msg(Info(name, tick, rate_in, wt_n, wt_i, None), wmsgs)
                        observer.on_next(res)

            # Determine Nc logic
            def on_next_Nc(x):
                if sync:
                    # Calculate expected number of message to be received
                    delay = params["delay"] if simulate_delays else 0.0
                    num_msgs = expected_inputs(x, rate_in, rate_node, delay, bool(skip))
                    num_queue.append(num_msgs)
                tick_queue.append(x)
                next(x)

            subscriptions = []
            sad = SingleAssignmentDisposable()
            sad.disposable = source_Nc.subscribe(on_next_Nc, observer.on_error, observer.on_completed, scheduler)
            subscriptions.append(sad)

            def on_next_msg(x):
                msgs_queue.append(x[1])
                wc = time.time()
                seq = x[0]
                if sync:
                    sc = round(x[0] * dt_i, 12)
                else:
                    sc = (wc - start) / real_time_factor
                t_i_queue.append(Stamp(seq, sc, wc))
                next(x)

            sad = SingleAssignmentDisposable()
            if not sync and simulate_delays:
                source_msg_delayed = source_msg.pipe(ops.delay(params["delay"] / real_time_factor))
            else:
                source_msg_delayed = source_msg
            sad.disposable = source_msg_delayed.subscribe(on_next_msg, observer.on_error, observer.on_completed, scheduler)
            subscriptions.append(sad)

            return CompositeDisposable(subscriptions)

        return rx.create(subscribe)

    return _generate_msgs


def create_channel(
    ns,
    Nc,
    rate_node,
    inpt,
    sync,
    real_time_factor,
    simulate_delays,
    E,
    scheduler,
    is_feedthrough,
    node,
):
    if is_feedthrough:
        name = inpt["feedthrough_to"]
    else:
        name = inpt["name"]

    # Readable format
    Is = inpt["reset"]
    Ir = inpt["msg"].pipe(
        convert(inpt["space"], inpt["processor"], name, "inputs", node, direction="in"),
        # ops.combine_latest(E),  # Throttle with end_reset
        # ops.map(lambda x: x[0]),  # Only pass through message
        ops.observe_on(scheduler),
        ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
        ops.share(),
    )

    # Get rate from rosparam server
    rate_str = "%s/rate/%s" % (ns, inpt["address"][len(ns) + 1 :])
    rate = eagerx.utils.utils.get_param_with_blocking(rate_str, node.backend)

    # Create input channel
    if real_time_factor == 0:
        Nc = Nc.pipe(ops.observe_on(scheduler), ops.start_with(0))
    else:
        Nc = Nc.pipe(ops.observe_on(scheduler))

    channel = Ir.pipe(
        generate_msgs(
            Nc,
            rate_node,
            name,
            rate,
            params=inpt,
            sync=sync,
            real_time_factor=real_time_factor,
            simulate_delays=simulate_delays,
            node=node,
        ),
        ops.share(),
    )

    # Create reset flag
    flag = Ir.pipe(
        ops.map(lambda val: val[0] + 1),
        ops.start_with(0),
        ops.combine_latest(Is),  # Depends on ROS reset msg type
        ops.filter(lambda value: not sync or value[0] == value[1]),
        ops.map(lambda x: {name: x[0]}),
    )
    return channel, flag


def init_channels(
    ns,
    Nc,
    rate_node,
    inputs,
    sync,
    real_time_factor,
    simulate_delays,
    E,
    scheduler,
    node,
    is_feedthrough=False,
):
    # Create channels
    channels = []
    flags = []
    for i in inputs:
        channel, flag = create_channel(
            ns,
            Nc,
            rate_node,
            i,
            sync,
            real_time_factor,
            simulate_delays,
            E,
            scheduler,
            is_feedthrough,
            node,
        )
        channels.append(channel)
        if is_feedthrough:
            name = i["address"]
        else:
            name = i["name"]
        flag_name = "flag [%s]" % name.split("/")[-1][:12].ljust(4)
        # flag.pipe(spy(f'sub_{flag_name}', node)).subscribe(print)
        flag = flag.pipe(spy(flag_name, node))  # , ops.take(1), ops.merge(rx.never()))
        flags.append(flag)
    zipped_flags = rx.zip(*flags).pipe(ops.map(lambda x: merge_dicts({}, x)))
    zipped_channels = rx.zip(*channels).pipe(
        ops.combine_latest(
            E.pipe(ops.observe_on(scheduler))
        ),  # Latch output on '/end_reset' --> Can only receive 1 each episode.
        ops.map(lambda x: x[0]),
        regroup_inputs(node, rate_node=rate_node),
        ops.share(),
    )
    return zipped_channels, zipped_flags


def init_real_reset(
    ns,
    Nc,
    rate_node,
    RR,
    real_reset,
    feedthrough,
    targets,
    sync,
    real_time_factor,
    simulate_delays,
    E,
    scheduler,
    node,
):
    # Create real reset pipeline
    dispose = []
    if real_reset:
        for i in feedthrough:
            rate_str = "%s/rate/%s" % (ns, i["address"][len(ns) + 1 :])
            rate_in = eagerx.utils.utils.get_param_with_blocking(rate_str, node.backend)
            if not rate_in == rate_node:
                raise ValueError(
                    "Rate of the reset node (%s) must be exactly the same as the feedthrough node rate (%s)."
                    % (rate_node, rate_in)
                )

        # Create zipped action channel
        zipped_channels, zipped_flags = init_channels(
            ns,
            Nc,
            rate_node,
            feedthrough,
            sync,
            real_time_factor,
            simulate_delays,
            E,
            scheduler,
            node,
            is_feedthrough=True,
        )

        # Create switch subject
        target_signal = rx.zip(*[t["msg"] for t in targets])
        RR_ho = BehaviorSubject(zipped_channels)
        d_RR_ho = RR.pipe(
            ops.combine_latest(target_signal),  # make switch logic wait for all targets to be received.
            ops.map(lambda x: Nc.pipe(ops.map(lambda x: None), ops.start_with(None))),
        ).subscribe(RR_ho)
        rr_channel = RR_ho.pipe(ops.switch_latest())

        # Add disposables
        dispose += [RR_ho, d_RR_ho]
    else:
        # Create switch Subject
        zipped_flags = rx.never().pipe(ops.start_with({}))
        rr_channel = Nc.pipe(ops.map(lambda x: None), ops.start_with(None))

    return rr_channel, zipped_flags, dispose


def init_target_channel(states, scheduler, node):
    channels = []
    for s in states:
        c = s["msg"].pipe(
            convert(s["space"], s["processor"], s["name"], "targets", node, direction="in"),
            ops.share(),
            ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
            remap_target(s["name"], node.sync, node.real_time_factor),
        )
        channels.append(c)
    # HACK!: Why do we sometimes receive the targets twice? And is the first received the new target or the old one?
    #        In this way, we risk reusing a target twice.
    return rx.zip(*channels).pipe(regroup_inputs(node, is_input=False), ops.take(1))


def merge_dicts(dict_1, dict_2):
    if isinstance(dict_2, dict):
        dict_2 = (dict_2,)
    for d in dict_2:
        dict_1.update(d)
    return dict_1


def init_state_inputs_channel(ns, state_inputs, scheduler, node):
    if len(state_inputs) > 0:
        channels = []
        for s in state_inputs:
            d = s["done"].pipe(
                ops.scan(lambda acc, x: x if x else acc, False),
            )
            c = s["msg"].pipe(
                convert(s["space"], s["processor"], s["name"], "states", node, direction="in"),
                ops.share(),
                ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
                ops.start_with((-1, None)),
                ops.combine_latest(d),
                ops.filter(lambda x: x[0][0] >= 0 or x[1]),
                remap_state(s["name"], node.sync, node.real_time_factor),
            )
            channels.append(c)
        return rx.zip(*channels).pipe(regroup_inputs(node, is_input=False), ops.merge(rx.never()))
    else:
        return rx.never().pipe(ops.start_with(dict()))


def call_state_reset(state):
    def _call_state_reset(source):
        def subscribe(observer, scheduler=None):
            def on_next(state_msg):
                try:
                    state.reset(state=state_msg.msgs[0])
                except Exception as e:
                    observer.on_error(e)
                observer.on_next(state_msg)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _call_state_reset


def init_state_resets(ns, state_inputs, trigger, scheduler, tp_scheduler, node):
    if len(state_inputs) > 0:
        channels = []

        for s in state_inputs:
            d = s["done"].pipe(
                ops.scan(lambda acc, x: x if x else acc, False),
            )
            c = s["msg"].pipe(
                convert(s["space"], s["processor"], s["name"], "states", node, direction="in"),
                ops.share(),
                ops.scan(lambda acc, x: (acc[0] + 1, x), (-1, None)),
                ops.start_with((-1, None)),
                ops.combine_latest(d),
                ops.filter(lambda x: x[0][0] >= 0 or x[1]),
                remap_state(s["name"], node.sync, node.real_time_factor),
            )

            done, reset = trigger.pipe(
                with_latest_from(c),
                ops.take(1),
                ops.merge(rx.never()),
                ops.map(lambda x: x[1]),
                ops.partition(lambda x: x.info.done),
            )
            reset = reset.pipe(ops.observe_on(tp_scheduler), call_state_reset(s["state"]), ops.observe_on(scheduler))
            rs = rx.merge(
                done.pipe(spy("done [%s]" % s["name"].split("/")[-1][:12].ljust(4), node)),
                reset.pipe(spy("reset [%s]" % s["name"].split("/")[-1][:12].ljust(4), node)),
            )

            channels.append(rs)
        return rx.zip(*channels).pipe(regroup_inputs(node, is_input=False), ops.merge(rx.never()))
    else:
        return rx.never().pipe(ops.start_with(dict()))


def init_callback_pipeline(
    ns,
    cb_tick,
    cb_ft,
    stream,
    real_reset,
    targets,
    state_outputs,
    outputs,
    scheduler,
    node,
):
    d_msg = []
    if real_reset:
        target_stream = init_target_channel(targets, scheduler, node)

        # Split stream into feedthrough (ft) and reset stream
        reset_stream, ft_stream = stream.pipe(ops.partition(lambda x: x[1][1] is None))

        # Either feedthrough action or run callback
        ft_stream = ft_stream.pipe(
            ops.map(lambda x: x[1][1]),
            spy("CB_FT", node, log_level=DEBUG, mapper=remap_cb_input(mode=0)),
            ops.map(lambda val: cb_ft(val, node.sync)),
            ops.share(),
        )
        reset_stream = reset_stream.pipe(
            ops.map(lambda x: x[1][0]),
            ops.combine_latest(target_stream),
            spy("CB_RESET", node, log_level=DEBUG, mapper=remap_cb_input(mode=0)),
            ops.map(lambda val: cb_tick(**val[0], **val[1])),
            ops.share(),
        )
        output_stream = rx.merge(reset_stream, ft_stream)

        # Send done flags
        for s in state_outputs:
            d = reset_stream.pipe(
                spy("reset", node, log_level=DEBUG),
                ops.pluck(s["name"] + "/done"),
                ops.share(),
            ).subscribe(s["msg"])

            # Add disposable
            d_msg += [d]
    else:
        output_stream = stream.pipe(
            ops.filter(lambda x: x[1][1] is None),
            ops.map(lambda x: x[1][0]),
            spy("CB_TICK", node, log_level=DEBUG, mapper=remap_cb_input(mode=0)),
            ops.map(lambda val: cb_tick(**val)),
            ops.share(),
        )
    return d_msg, output_stream


def switch_with_check_pipeline(init_ho=None):
    if init_ho is None:
        stream_ho = Subject()
    else:
        stream_ho = BehaviorSubject(init_ho)
    check, stream = stream_ho.pipe(ops.switch_latest(), ops.partition(lambda event: event is None))
    return check, stream, stream_ho


def node_reset_flags(ns, node_flags, node):
    flags = [
        nf["msg"].pipe(
            flag_dict(nf["name"]),
            spy("has_reset", node),
            ops.start_with({nf["name"]: False}),
        )
        for nf in node_flags
    ]
    init_dict = dict()
    for nf in node_flags:
        init_dict[nf["name"]] = False
    stream = rx.merge(*flags).pipe(
        ops.scan(lambda acc, x: merge_dicts(acc, x), init_dict),
        ops.filter(lambda x: any([value for key, value in x.items()])),
        filter_dict(),
        spy("awaiting", node, log_level=DEBUG),
        ops.filter(lambda x: len(x) == 0),
        ops.start_with(None),
    )
    return stream


def filter_dict_on_key(key):
    def _filter_dict_on_key(source):
        def subscribe(observer, scheduler=None):
            def on_next(value):
                if key in value:
                    observer.on_next(value[key])

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _filter_dict_on_key


def throttle_with_time(dt, node, rate_tol: float = 0.95, log_level: int = DEBUG):
    time_fn = lambda: time.monotonic_ns() / 1e9  # noqa: E731
    node_name = node.ns_name
    color = node.color
    effective_log_level = node.backend.log_level
    log_time = 2  # [s]

    def _throttle_with_time(source):
        def subscribe(observer, scheduler=None):
            # Timing
            tic = [None]
            cum_cbs = [0]
            cum_delay = [0]
            cum_sleep = [0]

            # Logging
            last_cum_cbs = [0]
            last_cum_delay = [0]
            last_cum_sleep = [0]
            last_Nc = [0]
            last_time = [None]

            def on_next(value):
                Nc, start = value
                if Nc == 0:  # Do not throttle before the first callback
                    # NOTE: This is the first time we receive a value
                    sleep_time = 0.0
                else:  # Determine sleep time
                    assert tic[0] is not None, "tic is None"
                    toc = time_fn()
                    dt_comp = toc - tic[0]
                    sleep_time = dt - dt_comp  # if sleep_time > 0 then we are early, if sleep_time < 0 then we are late

                # Throttle callback
                if sleep_time > 0:  # Sleep if we are early
                    time.sleep(sleep_time)
                    cum_sleep[0] += sleep_time
                else:  # If we are overdue, the proceeed
                    cum_delay[0] += -sleep_time
                    cum_cbs[0] += 1
                tic[0] = start + Nc / node.rate
                # node.backend.loginfo(colored(f"[{node_name}] Nc: {Nc} | toc: {toc} | sleep_time: {sleep_time} | tic[0]: {tic[0]}", color))

                # Logging
                curr = time_fn()
                last_time[0] = last_time[0] if last_time[0] is not None else curr
                if (curr - last_time[0]) > log_time:
                    # Calculate statistics since last logged instance
                    log_window = curr - last_time[0]
                    log_cbs = cum_cbs[0] - last_cum_cbs[0]
                    # log_delay = cum_delay[0] - last_cum_delay[0]
                    log_sleep = cum_sleep[0] - last_cum_sleep[0]
                    log_Nc = Nc - last_Nc[0]

                    # Log statistics if not keeping rate
                    Nc_expected = log_window / dt  # [ticks]
                    rate_ratio = log_Nc / Nc_expected
                    cbs_ratio = log_cbs / log_Nc
                    sleep_ratio = log_sleep / log_window
                    # delay_ratio = log_delay / log_window
                    if rate_ratio < rate_tol and node.log_level >= effective_log_level and WARN >= effective_log_level:
                        print_str = f"Running at {rate_ratio*100:.2f}% of rate ({1/dt} Hz) | {sleep_ratio*100:.2f}% sleep | {100 - sleep_ratio*100:.2f}% computation | {cbs_ratio*100: .2f}% callbacks delayed |"
                        print_info(
                            node_name,
                            "red",
                            f"last {log_window:.2f} s",
                            trace_type="",
                            value=print_str,
                            log_level=INFO,
                        )
                    elif node.log_level >= effective_log_level and log_level >= effective_log_level:
                        print_str = f"Running at {rate_ratio*100:.2f}% of rate ({1/dt} Hz) | {sleep_ratio*100:.2f}% sleep | {100 - sleep_ratio*100:.2f}% computation | {cbs_ratio*100: .2f}% callbacks delayed |"
                        print_info(
                            node_name,
                            color,
                            f"last {log_window:.2f} s",
                            trace_type="",
                            value=print_str,
                            log_level=log_level,
                        )

                    # Set baseline statistics
                    last_time[0] = curr
                    last_cum_cbs[0] = cum_cbs[0]
                    last_cum_delay[0] = cum_delay[0]
                    last_cum_sleep[0] = cum_sleep[0]
                    last_Nc[0] = Nc

                # Send tick for next callback
                observer.on_next(Nc)

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _throttle_with_time


def throttle_callback_trigger(rate_node, Nc, E, sync, real_time_factor, scheduler, node):
    # return Nc
    if sync and real_time_factor == 0:
        Nct = Nc
    else:
        assert (
            real_time_factor > 0
        ), "The real_time_factor must be larger than zero when *not* running reactive (i.e. asychronous)."
        wc_dt = 1 / (rate_node * real_time_factor)
        Nct = Nc.pipe(
            ops.scan(lambda acc, x: acc + 1, 0),
            ops.start_with(0),
            ops.combine_latest(E),  # .pipe(spy("Nct_E", node, log_level=INFO))),
            ops.observe_on(scheduler),
            # spy("Nct_E_CBL", node, log_level=INFO),
            throttle_with_time(wc_dt, node),
            ops.share(),
        )
    return Nct


def with_latest_from(*sources: Observable):
    def _with_latest_from(parent: Observable) -> Observable:
        NO_VALUE = NotSet()

        def subscribe(observer, scheduler=None):
            def subscribe_all(parent, *children):
                parent_queued = [None]
                values = [NO_VALUE for _ in children]

                def subscribe_child(i, child):
                    subscription = SingleAssignmentDisposable()

                    def on_next(value):
                        with parent.lock:
                            values[i] = value
                            if parent_queued[0] is not None and NO_VALUE not in values:
                                result = (parent_queued[0],) + tuple(values)
                                parent_queued[0] = None
                                observer.on_next(result)

                    subscription.disposable = child.subscribe_(on_next, observer.on_error, scheduler=scheduler)
                    return subscription

                parent_subscription = SingleAssignmentDisposable()

                def on_next(value):
                    with parent.lock:
                        if NO_VALUE not in values:
                            result = (value,) + tuple(values)
                            observer.on_next(result)
                        else:
                            parent_queued[0] = value

                disp = parent.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
                parent_subscription.disposable = disp

                children_subscription = [subscribe_child(i, child) for i, child in enumerate(children)]

                return [parent_subscription] + children_subscription

            return CompositeDisposable(subscribe_all(parent, *sources))

        return Observable(subscribe)

    return _with_latest_from


class NotSet:
    """Sentinel value."""

    def __eq__(self, other):
        return self is other

    def __repr__(self):
        return "NotSet"


def convert(space: eagerx.Space, processor, name, component, node, direction="out"):
    OUTPUT = True if direction == "out" else False
    INPUT = True if direction == "in" else False
    space_checked = [False]
    p_msg = f" (after processing with `{processor.__class__.__qualname__}`)" if processor else ""
    assert isinstance(space, eagerx.Space), f"The space of '{name}' is not of type eagerx.Space."

    def _convert(source):
        def subscribe(observer, scheduler=None):
            def on_next(recv):
                if INPUT:
                    # Preprocess input message
                    if processor is not None:
                        recv = processor.convert(recv)

                    # Check if message complies with space (after conversion)
                    if not space_checked[0]:
                        space_checked[0] = True
                        if not space.contains(np.array(recv)):
                            shape_msg = f"(msg.shape={recv.shape} vs space.shape={space.shape})"
                            dtype_msg = f"(msg.dtype={recv.dtype} vs space.dtype={space.dtype})"
                            msg = (
                                f"[subscriber][{node.ns_name}][{component}][{name}]: Message{p_msg} does not match the defined space. "
                                f"Either a mismatch in expected shape {shape_msg}, dtype {dtype_msg}, and/or the value is out of bounds (low/high)."
                            )
                            node.backend.logwarn_once(msg, identifier=f"[{node.ns_name}][{name}]")
                    observer.on_next(recv)
                elif OUTPUT:
                    # Convert python native types to numpy arrays.
                    if isinstance(recv, float):
                        recv = np.array(recv, dtype="float32")
                    elif isinstance(recv, bool):
                        recv = np.array(recv, dtype="bool")
                    elif isinstance(recv, int):
                        recv = np.array(recv, dtype="int64")

                    # Process message
                    if processor is not None:
                        recv = processor.convert(recv)
                    else:
                        recv = recv

                    if not space_checked[0]:
                        space_checked[0] = True
                        if not space.contains(np.array(recv)):
                            shape_msg = f"(msg.shape={recv.shape} vs space.shape={space.shape})"
                            dtype_msg = f"(msg.dtype={recv.dtype} vs space.dtype={space.dtype})"
                            msg = (
                                f"[publisher][{node.ns_name}][{component}][{name}]: Message{p_msg} does not match the defined space. "
                                f"Either a mismatch in expected shape {shape_msg}, dtype {dtype_msg}, and/or the value is out of bounds (low/high)."
                            )
                            node.backend.logwarn_once(msg, identifier=f"[{node.ns_name}][{name}]")

                    observer.on_next(recv)
                else:
                    raise NotImplementedError(f"Direction not implemented: {direction}.")

            return source.subscribe(on_next, observer.on_error, observer.on_completed, scheduler)

        return rx.create(subscribe)

    return _convert



import abc
import numpy as np
import typing

from eagerx.utils.utils import Header

if typing.TYPE_CHECKING:
    from eagerx.core.entities import Backend


class Publisher(object):
    def __init__(self, backend: "Backend"):
        self._backend = backend
        self._seq = 0

    @abc.abstractmethod
    def _publish(self, msg: typing.Union[float, bool, int, str, np.ndarray, np.number], header: Header) -> None:
        pass

    def publish(
        self, msg: typing.Union[float, bool, int, str, np.ndarray, np.number], header: typing.Optional[Header] = None
    ) -> None:
        if header is None:
            sc, wc = self._backend.now()
            header = Header(seq=self._seq, sc=sc, wc=wc)
        else:
            header._replace(seq=self._seq)
        self._publish(msg, header)

    @abc.abstractmethod
    def unregister(self) -> None:
        pass


class Subscriber(object):
    def __init__(self, backend: "Backend", header: bool):
        self._backend = backend
        self._header = header

    @abc.abstractmethod
    def unregister(self) -> None:
        pass


class ShutdownService(object):
    @abc.abstractmethod
    def unregister(self) -> None:
        pass



__version__ = "0.1.39"

from eagerx.core.space import Space
from eagerx.core.constants import (  # noqa: F401  # pylint: disable=unused-import
    process,
    log,
    SILENT,
    DEBUG,
    INFO,
    WARN,
    ERROR,
    FATAL,
    NEW_PROCESS,
    ENVIRONMENT,
    ENGINE,
    EXTERNAL,
)

from eagerx.core.entities import (  # noqa: F401  # pylint: disable=unused-import
    Object,
    Engine,
    Node,
    ResetNode,
    Processor,
    EngineState,
    EngineNode,
    Backend,
)
from eagerx.core.env import BaseEnv  # noqa # pylint: disable=unused-import
from eagerx.core.graph import Graph  # noqa # pylint: disable=unused-import
from eagerx.core.graph_engine import EngineGraph  # noqa # pylint: disable=unused-import
import eagerx.core.register as register  # noqa # pylint: disable=unused-import
import eagerx.core.specs as specs  # noqa: F401  # pylint: disable=unused-import
import eagerx.wrappers as wrappers  # noqa: F401  # pylint: disable=unused-import
from eagerx.core.log import (  # noqa: F401  # pylint: disable=unused-import
    get_log_level,
    set_log_level,
    logdebug,
    loginfo,
    logwarn,
    logerr,
    logfatal,
)



import typing as t
import gymnasium.spaces as spaces
import numpy as np
from functools import reduce
import operator as op


class Space(spaces.Space):
    """
    A (possibly unbounded) space in R^n. Specifically, a Space represents the
    Cartesian product of n closed intervals. Each interval has the form of one
    of [a, b], (-oo, b], [a, oo), or (-oo, oo).

    There are two common use cases:

    * Identical bound for each dimension::
        >>> Space(low=-1.0, high=2.0, shape=(3, 4), dtype="float32")
        Space(3, 4)

    * Independent bound for each dimension::
        >>> Space(low=np.array([-1.0, -2.0]), high=np.array([2.0, 4.0]), dtype="float32")
        Space(2,)

    """

    def __init__(
        self,
        low: t.Optional[t.Union[t.List, np.ndarray, int, float, bool, t.Tuple]] = None,
        high: t.Optional[t.Union[t.List, np.ndarray, int, float, bool, t.Tuple]] = None,
        shape: t.Optional[t.Tuple] = None,
        dtype: t.Union[np.number, str] = np.float32,
        seed: t.Optional[int] = None,
    ):
        """Constructor for initializing a space.

        :param low: Expected lower bound of messages.
        :param high: Expected upper bound of messages.
        :param shape: Expected shape of messages.
        :param dtype: Expected datatype of messages.
        :param seed: Seed that is used to draw random samples from the space.
        """
        if isinstance(low, (list, tuple)):
            low = np.array(low, dtype=dtype)
        if isinstance(high, (list, tuple)):
            high = np.array(high, dtype=dtype)

        # determine shape if it isn't provided directly
        if shape is not None:
            shape = tuple(shape)
            assert low is None or np.isscalar(low) or low.shape == shape, "low.shape doesn't match provided shape"
            assert high is None or np.isscalar(high) or high.shape == shape, "high.shape doesn't match provided shape"
        elif low is not None and not np.isscalar(low):
            shape = low.shape
            assert high is not None and np.isscalar(high) or high.shape == shape, "high.shape doesn't match low.shape"
        elif high is not None and not np.isscalar(high):
            shape = high.shape
            assert low is not None and np.isscalar(low) or low.shape == shape, "low.shape doesn't match high.shape"
        elif shape is None:
            pass
        else:
            raise ValueError("shape must be provided or inferred from the shapes of low or high")

        if np.isscalar(low):
            low = np.full(shape, low, dtype=dtype)

        if np.isscalar(high):
            high = np.full(shape, high, dtype=dtype)

        if low is not None and high is not None:
            self._space = spaces.Box(low=low, high=high, shape=shape, dtype=dtype, seed=seed)
        else:
            self._low = low
            self._high = high
            self._space = None
        super().__init__(shape=shape, dtype=dtype, seed=seed)

    def __getattr__(self, name):
        if self._space is None:
            raise AttributeError("Attribute does not exist.")
        else:
            return getattr(self._space, name)

    def contains(self, x) -> bool:
        """
        Return boolean specifying if x is a valid
        member of this space
        :param x: array to check.
        """
        if self._space is None:
            # todo: convert to ndarray? Maybe not to test that dtype is correct && we are dealing with np array?
            # if not isinstance(x, np.ndarray):
            #     x = np.asarray(x, dtype=self.dtype)

            return x.dtype == self.dtype and (self.shape is None or x.shape == self.shape)
        else:
            return x.dtype == self.dtype and self._space.contains(x)  # Check dtype separately

    def contains_space(self, space: t.Union["Space", t.Dict]) -> bool:
        """
        Return boolean specifying if space is contained
        in this space.
        Low and high of the space must exactly match (instead of lying within the bounds) to return True
        :param space: Space that is to be checked.
        """
        if isinstance(space, dict):
            space = Space.from_dict(space)

        # Check dtype
        flag = space.dtype == self.dtype

        # Check shape
        if flag and space.shape is not None and self.shape is not None:
            flag = flag and space.shape == self.shape

        # Check bounds
        if flag and space._space is not None and self._space is not None:
            flag = flag and (space._space.low == self._space.low).all()
            flag = flag and (space._space.high == self._space.high).all()

        return flag

    def sample(self) -> np.ndarray:
        """Randomly sample an element of this space. Can be
        uniform or non-uniform sampling based on boundedness of space."""
        if not self.is_fully_defined:
            raise NotImplementedError("Cannot sample from this incomplete space (i.e. low, high not provided?).")
        else:
            return self._space.sample()

    def to_dict(self) -> t.Dict:
        """Convert the space to a dict representation

        :return: Dict representation of the space.
        """
        if self._space is None:
            if self.shape is None:
                return dict(low=None, high=None, shape=None, dtype=self.dtype.name)
            else:
                return dict(low=None, high=None, shape=list(self.shape), dtype=self.dtype.name)
        else:
            low = np.nan_to_num(self.low)
            high = np.nan_to_num(self.high)
            if np.all(low == low.reshape(-1)[0]):
                low = low.reshape(-1)[0].item()
            if np.all(high == high.reshape(-1)[0]):
                high = high.reshape(-1)[0].item()
            if isinstance(low, np.ndarray):
                low = low.tolist()
            if isinstance(high, np.ndarray):
                high = high.tolist()
            return dict(low=low, high=high, shape=list(self.shape), dtype=self.dtype.name)

    @property
    def is_np_flattenable(self):
        """Checks whether this space can be flattened."""
        return True

    @property
    def is_fully_defined(self) -> bool:
        """Check if space is fully defined (i.e. low, high, shape and dtype are all provided).
        :return: flag
        """
        return self._space is not None

    @classmethod
    def from_dict(cls, d: t.Dict) -> "Space":
        """Create a space from a dict.

        :param d: Dict containing the arguments to initialize the space
        :return: The space.
        """
        return cls(low=d["low"], high=d["high"], shape=d["shape"], dtype=d["dtype"])

    def __repr__(self):
        if self._space is None:
            if self.shape is None:
                return f"Space({self.dtype})"
            else:
                return f"Space({self.shape}, {self.dtype})"
        else:
            return f"Space({self.low}, {self.high}, {self.shape}, {self.dtype})"

    def __eq__(self, other):
        if self._space is None:
            return (
                isinstance(other, Space)
                and other._space is None
                and (self.shape == other.shape)
                and (self.dtype == other.dtype)
            )
        else:
            return (
                isinstance(other, spaces.Box)
                and (self.shape == other.shape)
                and np.allclose(self.low, other.low)
                and np.allclose(self.high, other.high)
            )


# Register a method to flatten our custom Space using standard gym functions.
@spaces.utils.flatten_space.register(Space)
def flatten_space_box(space):
    return spaces.Box(space.low.flatten(), space.high.flatten(), dtype=space.dtype)


# Register a method to unflatten our custom Space using standard gym functions.
@spaces.utils.unflatten.register(Space)
def unflatten_box(space, x):
    return np.asarray(x, dtype=space.dtype).reshape(space.shape)


@spaces.utils.flatdim.register(Space)
def flatdim_box_multibinary(space):
    return reduce(op.mul, space.shape, 1)


@spaces.utils.flatten.register(Space)
def flatten_box_multibinary(space, x):
    return np.asarray(x, dtype=space.dtype).flatten()
